import os
from math import prod

import nibabel as nb

from .types import CheckList, CheckResult


def load(path: os.PathLike[str]) -> nb.Nifti1Image:
    """Load NIfTI-1 or NIfTI-2 (including CIFTI-2) images"""
    try:
        return nb.Nifti1Image.from_filename(path)
    except Exception:
        return nb.Nifti2Image.from_filename(path)


def fullsize(img: nb.Nifti1Image) -> CheckResult:
    """Verify NIfTI image has expected length from header"""
    dataobj: nb.arrayproxy.ArrayProxy = img.dataobj  # type: ignore

    expected: int = dataobj.offset + dataobj.dtype.itemsize * prod(dataobj.shape)

    with dataobj._get_fileobj() as fobj:
        # Seek beyond end returns end position
        actual: int = fobj.seek(expected + 1)
    return CheckResult(
        (actual == expected, f'Expected {expected} bytes; found {actual}')
    )


CHECKS = CheckList(
    loader=load,
    checks=[
        fullsize,
    ],
)



import json
import os
import re
import typing as ty
from enum import Enum
from pathlib import Path

import nibabel as nb
import typer
from typing_extensions import TypeAlias

from . import nifti
from .types import CheckList, FailedCheck

# Add to this list as file types are added
SupportedType: TypeAlias = ty.Union[nb.Nifti1Image,]

# Regex patterns paired with checklists to verify on file
# To avoid multiple reads, only the first matching check is applied to each
# file
CHECKS: list[tuple[re.Pattern[str], CheckList[SupportedType]]] = [
    (re.compile(r'\.nii(\.gz)?$'), nifti.CHECKS),
]

app = typer.Typer()


class AnalysisLevel(Enum):
    run = 'run'
    session = 'session'
    participant = 'participant'
    dataset = 'dataset'


@app.command()
def main(
    bids_dir: Path,
    output_dir: ty.Optional[Path] = typer.Argument(None),
    analysis_level: ty.Optional[AnalysisLevel] = typer.Argument(None),
    allow_dangling_links: bool = typer.Option(
        False,
        '--allow-dangling-links',
        '-l',
        help='Ignore symlinks with missing targets',
    ),
) -> None:
    errors = []
    for root, dirs, files in os.walk(bids_dir):
        dirs[:] = [d for d in dirs if not d[0] == '.']
        root_path = Path(root)
        for file in files:
            for pattern, checklist in CHECKS:
                if re.search(pattern, file):
                    path = root_path / file
                    try:
                        fileobj = checklist.loader(path)
                        for check in checklist.checks:
                            result, message = check(fileobj)
                            if not result:
                                raise FailedCheck(message or check.__doc__)
                    except Exception as e:
                        if not allow_dangling_links or not isinstance(
                            e, FileNotFoundError
                        ):
                            errors.append(
                                (
                                    str(path.relative_to(bids_dir)),
                                    type(e).__name__,
                                    str(e),
                                )
                            )
                    break
    print(json.dumps(errors, indent=2))
    raise typer.Exit(code=len(errors) > 0)


if __name__ == '__main__':
    app()



import os
import typing as ty
from dataclasses import dataclass

T = ty.TypeVar('T')

CheckResult = ty.NewType('CheckResult', tuple[bool, str])


@dataclass
class CheckList(ty.Generic[T]):
    """Pair a file loader with boolean checks on the loaded file"""

    loader: ty.Callable[[os.PathLike[str]], T]
    checks: list[ty.Callable[[T], CheckResult]]


class FailedCheck(Exception):
    pass



import json
from pathlib import Path

import pytest
from typer.testing import CliRunner

from miniqc.__main__ import app

runner = CliRunner()


@pytest.mark.parametrize('allow_dangling', (True, False))
def test_app(example_dataset: Path, allow_dangling: bool) -> None:
    args = ['-l', str(example_dataset)]
    if not allow_dangling:
        args = args[1:]

    result = runner.invoke(app, args)

    assert result.exit_code == 1
    errors: list[list[str]] = json.loads(result.stdout)
    assert len(errors) == 1 if allow_dangling else 2
    assert [
        'sub-01/anat/sub-01_acq-truncated_T2w.nii.gz',
        'FailedCheck',
        'Expected 477 bytes; found 352',
    ] in errors

    if not allow_dangling:
        assert any([err[1] == 'FileNotFoundError' for err in errors])
